# Ksplice Uptrack API client library for Python
#
# Copyright (C) 2010 Ksplice, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import cjson
import posixpath
import ConfigParser
import pycurl
# Special handling for old versions of pycurl
try:
    pycurl.E_COULDNT_RESOLVE_PROXY
except:
    ### Comment copied from uptrack-upgrade ###
    # Values from an enum in include/curl/curl.h.  Comments in that
    # header say these should never change, so it seems reasonably
    # safe to use them on old versions of pycurl where these constants
    # are not exported to Python.
    pycurl.E_COULDNT_RESOLVE_PROXY = 5
    pycurl.E_COULDNT_RESOLVE_HOST = 6
    pycurl.E_COULDNT_CONNECT = 7
    pycurl.E_OPERATION_TIMEOUTED = 28
    pycurl.E_URL_MALFORMAT = 3
    pycurl.E_SSL_CACERT = 60

from StringIO import StringIO

from ksplice.uptrack.errors import UptrackError, AuthenticationError
from ksplice.uptrack.errors import UsageError, NetworkError

API_URL         = "https://uptrack.ksplice.com/api/"
API_VERSION     = 1
API_CONFIG_FILE = "/etc/uptrack-api.conf"
BUG_EMAIL       = "devel@ksplice.com"

class Client(object):
    """ A client for the Ksplice Uptrack API."""
    def __init__(self, user=None, api_key=None, api_url=None,
                 proxy=None, proxy_port=None):
        """
        Accepts a username and API key for authentication. Optionally,
        give a hostname and port for an HTTP proxy to use for
        requests.

        If the username or api key are not provided, they will be
        retrieved from the environment variables
        `UPTRACK_API_USERNAME` and `UPTRACK_API_KEY`.

        Finally, if a username and API key still cannot be found, the
        file `/etc/uptrack-api.conf` will be loaded. This file should
        be of the form:

        [uptrack]
        username = joeuser
        api_key  = 3af3c2c1ec407feb0fdc9fc1d8c4460c

        Your username and API key can be retrieved from
        <https://uptrack.ksplice.com/settings>.

        Similarly, if `proxy` is not set, the `https_proxy` and
        `HTTPS_PROXY` environment variables are checked, followed by
        `/etc/uptrack-api.conf`. If `proxy` is set and `proxy_port` is
        not set, `proxy` contains the port information, eg.
        `http://proxy.example.com:3128/`.
        """
        if user is None:
            user = os.environ.get('UPTRACK_API_USERNAME')
        if api_key is None:
            api_key = os.environ.get('UPTRACK_API_KEY')
        if api_url is None:
            api_url = os.environ.get('UPTRACK_API_URL', None)
        if proxy is None:
            proxy = os.environ.get('https_proxy')
        if proxy is None:
            proxy = os.environ.get('HTTPS_PROXY')

        config = ConfigParser.SafeConfigParser()
        try:
            config.read([API_CONFIG_FILE])
            if user is None and config.has_option('uptrack', 'username'):
                user = config.get('uptrack', 'username')
            if api_key is None and config.has_option('uptrack', 'api_key'):
                api_key = config.get('uptrack', 'api_key')
            if api_url is None and config.has_option('uptrack', 'api_url'):
                api_url = config.get('uptrack', 'api_url')
            if proxy is None and config.has_option('uptrack', 'https_proxy'):
                proxy = config.get('uptrack', 'https_proxy')
        except ConfigParser.Error, e:
            raise UsageError(e.message)

        if user is None or api_key is None:
            raise UsageError("You must provide an API username and key.")

        if api_url is None:
            api_url = API_URL

        self.user = user
        self.api_key = api_key
        self.api_url = api_url

        self.proxy = proxy
        self.proxy_port = proxy_port

        self.connection = self.initCurl()

    def request(self, method, path, data=None):
        """Make an HTTP request to the Uptrack API.

        `method` - The HTTP method to use
        `path`   - The path on the server to request
        `data`   - Data to include (JSON-encoded) as the request body

        On a successful request (HTTP code 2xx), returns the
        JSON-decoded body of the server's response, or raises an
        AuthenticationError or UptrackError on error.
        """
        if method == 'GET':
            self.connection.setopt(pycurl.HTTPGET, 1)
        elif method == 'POST':
            self.connection.setopt(pycurl.POST, 1)
            if data is not None:
                body = cjson.encode(data)
            else:
                body = ''
            self.connection.setopt(pycurl.POSTFIELDS, body)
        else:
            msg = "The %s HTTP method is not supported." % (method,)
            raise NotImplementedError(msg)

        full_path = posixpath.join(self.api_url,
                                   str(API_VERSION),
                                   path)
        self.connection.setopt(pycurl.URL, full_path)

        if self.proxy:
            self.connection.setopt(pycurl.PROXY, self.proxy)
        if self.proxy_port:
            self.connection.setopt(pycurl.PROXYPORT, self.proxy_port)

        pageContents = StringIO()
        self.connection.setopt(pycurl.WRITEFUNCTION, pageContents.write)
        try:
            self.connection.perform()
        except pycurl.error, e:
            raise NetworkError(self.resultFromPycurl(e))

        status = self.connection.getinfo(pycurl.HTTP_CODE)

        body = pageContents.getvalue()

        if status < 200 or status > 299:
            if status == 403:
                raise AuthenticationError(403, body)
            else:
                raise UptrackError(status, body)

        if len(body):
            return cjson.decode(body)
        return None

    def list_machines(self):
        """ List all machines known by the Uptrack service.

        Returns a list of dictionaries of the form:

         {
            'hostname': 'uptrack.ksplice.com',
            'ip': '184.73.248.238',
            'last_seen': '2010-04-26T18:03:43Z',
            'uuid': 'e82ba0ae-ad0a-4b92-a776-62b502bfd29d',
            'status': 'uptodate',
            'active': True
         }

        `status` will be one of `unsupported`, `outofdate`, and `uptodate`.
        """
        return self.request('GET', 'machines')

    def describe_machine(self, uuid):
        """ Obtain details about a specific machine by its UUID.

        If the account has a machine with the given UUID, this function will
        return a dictionary of the form:

         {
            'hostname': 'uptrack.ksplice.com',
            'ip': '184.73.248.238',
            'last_seen': '2010-04-26T18:03:43Z',
            'uuid': 'e82ba0ae-ad0a-4b92-a776-62b502bfd29d',
            'status': 'uptodate',
            'active': True,
            'steps': [['Install', {'ID': 'yqcei7u6',
                                   'Name': 'Clear garbage data on the kernel stack when handling signals.'}],
                      ['Install', {'ID': '0w8npm73',
                                   'Name': 'Kernel crash forwarding network traffic.'}],
                      ['Remove',  {'ID': 'diptbg4f',
                                   'Name': 'CVE-2010-0415: Information Leak in sys_move_pages'}],
                      ['Install', {'ID': 'xd3qw6mj',
                                   'Name': 'Floating point state corruption after signal.'}]]
         }

        `status` will be one of `unsupported`, `outofdate`, and `uptodate`.

        `steps` is the list of steps that need to be taken to bring
        the machine up to date.  Each step is an update to install or
        remove.  It will normally only contain "Install" steps, but
        may occasionally include "Remove" lines, if, for example, an
        update has been superceded by a newer version.

        If the account does not have a machine with the given UUID, the
        request will result in a HTTP 404 error, which will raise an
        UptrackError.

        The Uptrack UUID of a machine is stored at the path
        /var/lib/uptrack/uuid.
        """
        return self.request('GET', posixpath.join('machine', uuid, 'describe'))

    def authorize_machine(self, uuid):
        """ Authorize a specific UUID to access the Uptrack service.

        If you enable the "manual authorization" setting in the
        Uptrack service, the service will accept requests using your
        access key only from machines whose UUID you have specifically
        authorized.  This API call authorizes a machine for the
        Uptrack service.

        You can enable the "manual authorization" setting at
        https://uptrack.ksplice.com/settings/.

        The Uptrack UUID of a machine is stored at the path
        /var/lib/uptrack/uuid.
        """
        self.request('POST',
                     posixpath.join('machine', uuid, 'authorize'),
                     {'authorized': True})

    def deauthorize_machine(self, uuid):
        """ Revoke a machine's access to the Uptrack service.

        See `authorize_machine` for more information.
        """
        self.request('POST',
                     posixpath.join('machine', uuid, 'authorize'),
                     {'authorized': False})

    def initCurl(self):
        """ Initialize a cURL object with basic properties."""
        headers = ['X-Uptrack-User: ' + self.user,
                   'X-Uptrack-Key: ' + self.api_key,
                   'Content-Type: application/json']
        conn = pycurl.Curl()
        conn.setopt(pycurl.FOLLOWLOCATION, 1)
        conn.setopt(pycurl.MAXREDIRS, 5)
        conn.setopt(pycurl.CONNECTTIMEOUT, 30)
        conn.setopt(pycurl.TIMEOUT, 600)
        conn.setopt(pycurl.HTTPHEADER, headers)

        return conn

    def resultFromPycurl(self, error):
        """ Generate client-ready responses for network errors.

        `error` - A `pycurl.error`.
        """
        if error[0] in [pycurl.E_COULDNT_RESOLVE_HOST,
                        pycurl.E_COULDNT_CONNECT,
                        pycurl.E_OPERATION_TIMEOUTED]:
            msg = ("Could not connect to the Ksplice Uptrack server. "
                   "Check your network connection and try again.")
        elif error[0] == pycurl.E_COULDNT_RESOLVE_PROXY:
            msg = ("Could not resolve your proxy server while trying to "
                   "connect to the Ksplice Uptrack server.")
        elif error[0] == pycurl.E_URL_MALFORMAT:
            msg = ("Malformed URL <%s> for the Ksplice Uptrack server."
                   % (self.connection.getinfo(pycurl.EFFECTIVE_URL),))
        elif error[0] == pycurl.E_SSL_CACERT:
            msg = ("Could not verify the Ksplice Uptrack server's SSL "
                   "certificate.")
        else:
            msg = ("Unexpected error communicating with the Ksplice Uptrack "
                   "server.\nCheck your network connection and try again.")
        msg += ("\nPlease consult %s for assistance if you are unable to "
                "resolve this error." % (BUG_EMAIL,))
        return msg
